<?php

namespace Ems\Services;

use Doctrine\ORM\EntityManager;

use Util\SendEmail\SendEmail;
use Util\Environment;
use Util\Messages;

use Ems\Services\EstruturaService;
use Ems\Services\EscalaService;

use Ems\Repository\VinculoRepository;
use Ems\Repository\MarcacaoRepository;
use Ems\Repository\PeriodoRepository;
use Ems\Repository\EstruturaRepository;
use Ems\Util\{ Utilities, ArrayUtil, Time };

class VinculoService {
    
    const   TIPO_VINC_COLUMN    = 'NRTIPOVINCULOM';
    const   TIPO_VINC_CLT       = 1;
    const   TIPO_VINC_EST       = 2;
    
    const   GETHORARIOBASE_THROW_EXCEPTION          = true;
    const   GETHORARIOBASE_NOT_THROW_EXCEPTION      = false;
    const   GETHORARIOBASE_THROW_EXCEPTION_DEFAULT  = self::GETHORARIOBASE_NOT_THROW_EXCEPTION;

    /** @var EntityManager */
    protected $entityManager;

    /** @var Environment */
    protected $environment;

    /** @var VinculoRepository */
    protected $vinculoRepository;

    /** @var VinculoRepository */
    protected $periodoRepository;

    /** @var EstruturaService */
    protected $estruturaService;

    /** @var MarcacaoRepository */
    protected $marcacaoRepository;

    public function __construct(
        EntityManager      $entityManager,
        Environment        $environment,
        VinculoRepository  $vinculoRepository,
        PeriodoRepository  $periodoRepository,
        EstruturaService   $estruturaService,
        MarcacaoRepository $marcacaoRepository
    ) {
        $this->entityManager            = $entityManager;
        $this->environment              = $environment;
        $this->vinculoRepository        = $vinculoRepository;
        $this->periodoRepository        = $periodoRepository;
        $this->estruturaService         = $estruturaService;
        $this->marcacaoRepository       = $marcacaoRepository;
    }

    public function getVinculoHistorico(int $nrvinculom) {
        $nrorg = $this->environment->getNrorg();

        return $this->vinculoRepository->getVinculoHistorico($nrorg, $nrvinculom);
    }

    public function getVinculoMestreHistoricoByNrparcnegocio($nrparcnegocio) {
        $nrorg = $this->environment->getNrorg();

        return $this->vinculoRepository->getVinculoMestreHistoricoByNrparcnegocio($nrorg, $nrparcnegocio);
    }

    public function getVinculosGestor(
        int $nrEstrutura,
        $nrPeriodo = -1,
        $nrTipoVinculo = 1
    ): array {

        $nrOrg = $this->environment->getNrOrg();

        $periodo = null;

        if ($nrPeriodo === -1) {
            $periodoParams = [
                'P_NRORG'         => $nrOrg,
                'P_NRTIPOVINCULOM' => $nrTipoVinculo
            ];

            $periodo = $this->periodoRepository->getPeriodoAtual($periodoParams);
        } else {
            $periodo = $this->periodoRepository->getPeriodo($nrOrg, $nrPeriodo);
        }

        $mesDeCompetencia = Utilities::formatDate($periodo['DTFINALAPURACAO']);

        $estruturasInferioresEAtual = [];
        $this->estruturaService->
            getEstruturasInferioresRec($nrEstrutura, $estruturasInferioresEAtual, true);

        $estruturas = ArrayUtil::toStringSQL($estruturasInferioresEAtual);

        $params = [
            'P_NRORG'          => $nrOrg,
            'P_DTMESCOMPETENC' => $mesDeCompetencia,
            'P_NRESTRUTURAS'   => $estruturas
        ];

        $vinculos = $this->marcacaoRepository->getVinculosGestor($params);

        return $vinculos;
    }
    
    /**
     *  Função que retorna dados úteis do vínculo. 
     * 
     *  @param string $nrvinculom
     * 
     *  @return Array $dadosVinculo
     *      'NMVINCULOM'    => $nmvinculom
     * 
     *  @throws Exception
     */
    public function getDadosVinculo(
        string $nrvinculom,
        string $nrorg = null
    ) {
        if (empty($nrorg)) {
            $nrorg = $this->environment->getNrorg();
        }
        
        $vinculo = $this->vinculoRepository->getVinculoMestreHistorico($nrorg, $nrvinculom);
        
        if (empty($vinculo)) {
            throw new \Exception(Messages::VINC_NOT_FOUND);
        }
        
        return $vinculo;
    }
    
    /**
     *  Função que retorna o tipo vínculo. 
     * 
     *  @param string $tipoVinculo
     * 
     *  @return string  $
     * 
     *  @throws Exception
     */
    public function getTipoVinculo(
        string $nrvinculom,
        string $nrorg = null
    ) {
        if (empty($nrorg)) {
            $nrorg = $this->environment->getNrorg();
        }
        
        $vinculo = $this->getDadosVinculo($nrvinculom, $nrorg);
        
        $tipoVinculo = $vinculo[self::TIPO_VINC_COLUMN];
        
        return $tipoVinculo;
    }

    /**
     *  Rotina responsável por buscar a escala ativa de horários do funcionário e as horas a serem trabalhadas nessa escala. São retornados também informações sobre os horários do vínculo
     * 
     *  @param  int     $nrVinculo
     *  @param  bool    $throwException = false
     *  @param  string  $date = null
     * 
     *  @return $horarioBase (escala e horário)
     * 
     */
    public function getHorarioBase(
        int    $nrVinculo,
        string $date,
        bool   $throwException = self::GETHORARIOBASE_THROW_EXCEPTION_DEFAULT
    ): array {
        $nrOrg = (int) $this->environment->getNrorg();

        $horarioBase = $this->vinculoRepository->getHorarioBase($nrOrg, $nrVinculo, $date);

        if ($horarioBase === false) {
            if ($throwException) {
                throw new \Exception("Não foi possível encontrar o
                    horário base do vínculo $nrVinculo para o dia $date.");
            } else {
                return false;
            }
        }

        $qtHorasTrabalho = self::calculateHorasTrabalho($horarioBase);
        $horarioBase['QTHORASTRABALHO'] = $qtHorasTrabalho;

        $dadosHorario = $this->marcacaoRepository->getDadosHorarioVinculo(
            $nrVinculo, $nrOrg, $date
        );
        
        if (empty($dadosHorario)) {
            throw new \Exception(Messages::HOR_DADOS_NOT_FOUND);
        }

        $horarioBase = array_merge($dadosHorario, $horarioBase);

        return $horarioBase;
    }
    
    public function isHorarioEscalaTurno(
        int    $nrVinculo,
        string $date,
        bool   $throwException = VinculoService::GETHORARIOBASE_NOT_THROW_EXCEPTION
    ) {
        $escalaHorario = $this->getHorarioBase(
            $nrVinculo,
            $date,
            $throwException
        );
        
        if (empty($escalaHorario)) {
            return false;
        }
        
        return $escalaHorario['IDTIPOHORARIO'] == EscalaService::IDTIPOHORARIO_TURNO;
    }

    private static function calculateHorasTrabalho(
        array $horarioBase
    ): string {
        $primeiraEntrada = $horarioBase['PRIENTRADA'];
        $primeiraSaida   = $horarioBase['PRISAIDA'];

        $segundaEntrada = $horarioBase['SEGENTRADA'];
        $segundaSaida   = $horarioBase['SEGSAIDA'];

        $primeiraEntradaTime = Time::fromString($primeiraEntrada);
        $primeiraSaidaTime   = Time::fromString($primeiraSaida);

        $duracaoPrimeiroIntervaloTime = $primeiraSaidaTime->sub(
            $primeiraEntradaTime
        );

        $horasDeTrabalhoTime = null;

        // Estagiários não têm 2ª entrada e saída
        if ($segundaEntrada && $segundaSaida) {
            $segundaEntradaTime = Time::fromString($segundaEntrada);
            $segundaSaidaTime   = Time::fromString($segundaSaida);

            $duracaoSegundoIntervaloTime = $segundaSaidaTime->sub(
                $segundaEntradaTime
            );

            $horasDeTrabalhoTime = $duracaoPrimeiroIntervaloTime->sum(
                $duracaoSegundoIntervaloTime
            );
        } else {
            $horasDeTrabalhoTime = $duracaoPrimeiroIntervaloTime;
        }

        $horasDeTrabalho = $horasDeTrabalhoTime->toString();

        return $horasDeTrabalho;
    }

    public function getNrParcNegocioVinculo(
        int $nrVinculo
    ): int {
        $nrOrg = (int) $this->environment->getNrOrg();

        $params = [
            'P_NRORG'       => $nrOrg,
            'P_NRVINCULO'   => $nrVinculo,
        ];

        $nrParcNegocio = $this->vinculoRepository->getNrParcNegocioVinculo(
            $params
        );

        if (!$nrParcNegocio) {
            $message = 'O vínculo não possui número de parceiro de negócio!';
            throw new \Exception($message);
        }

        return (int) $nrParcNegocio;
    }
    
    public function getFeriadosVinculo(
        int     $nrVinculo,
        string  $dia
    ){
        $nrOrg = $this->environment->getNrorg();
        $dtmescompetenc = $this->periodoRepository->getDtmescompetencByPeriodo($nrOrg);
        
        return $this->vinculoRepository->isDiaFeriadoVinculo($nrOrg, $nrVinculo, $dtmescompetenc, $dia);
    }
    
    public function isDiaFeriadoVinculo(
        int     $nrVinculo,
        string  $dia
    ){
        $diaFeriadoVinculo = $this->getFeriadosVinculo($nrVinculo, $dia);
        return !empty($diaFeriadoVinculo);
    }
}
